/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AddressbooksSynchronizer.
 *
 * The Initial Developer of the Original Code is Günter Gersdorf.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Günter Gersdorf <G.Gersdorf@ggbs.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function doTest() {
  this.promptService.alert(null,'Test', 'doTest');
}

function initialize() {
  //this.doTest();
  var app = Services.appinfo;
  this.mAppStarttime = Math.round(Date.now()/1000);

  this.mVersion='';
  var abs=this;
  Components.utils.import("resource:///modules/AddonManager.jsm");
  AddonManager.getAddonByID("addressbookssync@ggbs.de",   //asynchron!!
    function(addOn) {
//        dump('AddOn: '+addOn+'\n');
//        for (var prop in addOn) {dump('Prop '+prop+': '+addOn[prop]+'\n');}
      abs.mVersion=addOn.version;
      abs.mPrefBranch.setCharPref('version',abs.mVersion);
      dump('AddressbooksSynchronizer '+abs.mVersion+' appName='+app.name+' appVersion='+app.version+'\n');
      Services.console.logStringMessage('AddressbooksSynchronizer '+abs.mVersion+
        ' appName='+app.name+' appVersion='+app.version+'\n');
    });

  this.update();  //Falls was zu updaten ist

  try { this.mNT = this.mPrefBranch.getBoolPref("NT"); } catch(e) { this.mNT = false; }

this.debug('App Start##############################################');
    //Used if !imap upload
    var observerService = Services.obs;
    observerService.addObserver(this,"quit-application-requested",false);
    observerService.addObserver(this,"quit-application-granted",false);
    observerService.addObserver(this,"domwindowclosed",false);
    //für true muß observer nsIWeakReference unterstützen
}

function finalize() {
  this.exiting=true;
  this.removeListener();  // remove the listener
this.debug('finalize: autoupload='+this.autoupload);
  if (this.autoupload) {
//if(this.dodebug)dump('onUnload autouploading:'+window.location+"\n");
this.debug('finalize: last window "'+this.lastWindow+'"');
    var app = Components.classes["@mozilla.org/toolkit/app-startup;1"].
        getService(Components.interfaces.nsIAppStartup);
    //definitely needs a window for remote/imap! (Also for local, or message to file fails)
    this.showPopup('upload', null, 'final', false);
//if(this.dodebug)dump('finalize popup shown:'+window.location+"\n");
this.debug('finalize: finalized');
  }
}

function update() {
// if something has to be updated (e.g. preferences)
}


/*
 * returns list of mab to be synced
 */
function syncedAddressbooks(direction, singleBook)
{   //direction: up, down, all
  var mabs=new Object();
  var mabsArray=Components.classes["@mozilla.org/supports-array;1"]
         .createInstance(Components.interfaces.nsICollection);

  var cn = this.allBooks;
  var found=false;
  while( cn.hasMoreElements() )
  {
    var dir = cn.getNext().QueryInterface(Components.interfaces.nsIAbDirectory);
    if (dir.dirType==this.kPABDirectory            // ignore ldap-addressbooks (dirType!=2)
          && this.getDB(dir)) {  // ignore non-Mozilla (eg. MACOs) books  //TODO: is there a better way?
      var sync=false;
      if (direction=='all')
        sync=true
      else if (singleBook && dir.fileName==singleBook)
        sync=true;
      else if (!singleBook) {
        var prefName = dir.dirPrefId;
        try { sync = this.mPrefBranch.getBoolPref(prefName+'.'+direction); }
          catch(e) { sync=false; }
      }
      if (sync) {
        found=true;
        var file = dir.fileName;
        if (mabs[file])
          // this ist a problem with the preferences, defining a book more than once!
          dump('allBooks returned double addressbook '+dir.fileName+'('+file+')\n');
        else {
          mabs[file]=true;
          mabsArray.AppendElement(dir);
        }
      }
    }
  } // while
  if (!found && direction=='down' && singleBook) { //its a new addressbook
    var sb=Components.classes["@mozilla.org/supports-string;1"]
        .createInstance(Components.interfaces.nsISupportsString);
    sb.data=singleBook;
    mabsArray.AppendElement(sb);
  }
  return mabsArray;
}

/*
 * return address book data structure associated with filename or description (=address book name)
 * TODO: might fail, if called with a mabName and the mabName is in a foreign language!
 *        (see mails with Ivan Kolchagov, to reproduce: change personal address book to bulgarien name)
 * called with a probably non-unique description only on manual downloads
 *   (LoadMab, DownloadMab) from options dialog
 */
function getAddressBook( name )
{
  var cp=name.indexOf('ldap_2.servers.')==0;
  var cn = this.allBooks;
  while( cn.hasMoreElements() )
  {
    var dir = cn.getNext().QueryInterface(Components.interfaces.nsIAbDirectory);
    var prefName = dir.dirPrefId;  //ldap_2.servers.xxx
    var description=dir.dirName;
    var filename;
    try {filename=dir.fileName;} catch(e) { filename='';} //might be 'not initialized'
//if(this.dodebug)dump('getAddressbook: compare '+name+' to '+prefName+', '+filename+', '+description+'\n');
    if( prefName == name || !cp && (filename == name || description == name)) {
        return dir;
    }
  } // while
  return null;
} // getAddressBook

/*
 * create new address book
 */
function createAddressBook(name, fileName)
{
  var uri=this.kMDBDirectoryRoot+fileName;
  try {
if(this.dodebug)dump(name+' create from '+uri+'\n');
//    this.directory.createDirectoryByURI(name, uri);
    var rooturi="moz-abdirectory://";
    this.abmanager.getDirectory(rooturi).createDirectoryByURI(name, uri);
    //for type see nsDirPrefs.h, PABDirectory=2
    //this ignores the uri
//    this.abmanager.newAddressBook(name, uri, 2/*type*/ /*,prefName*/);
  } catch (e) {
    this.promptService.alert(null, 'ERROR TB3', 'createAddressBook('+name+','+fileName+')\n'+e);
    return null;
  }
  var dir=this.getAddressBook(fileName);
  dir.lastModifiedDate=this.mAppStarttime; //set to 'not modified'
  return dir;
} // createAddressBook

/*
 * delete an address book
 */
function deleteAddressBook(dir)
{
  // Move focus away from the to be deleted addressbook
    // this has to be done in the AddressBook window :-((
//  try {  SetAbView(this.kPersonalAddressbookURI, false);} catch(e) {   this.promptService.alert(null,'ERROR', e);}

//  dir=dir.QueryInterface(Components.interfaces.nsIAbDirectory);
  try {
    var prefName = dir.dirPrefId;
    var mabFile=dir.fileName;
    var URI=dir.URI;
    if (URI==this.kPersonalAddressbookURI ||
        URI==this.kCollectedAddressbookURI)
      return false;

if(this.dodebug)dump('deleting '+URI+'\n');
//    this.directory.deleteDirectory(dir);
    this.abmanager.deleteAddressBook(URI);
      //prefName='ldap_2.servers.name' mit name='history', 'pab' oder eigener
    Services.prefs.clearUserPref( prefName+'.position' );  // sometimes remains!?
    Services.prefs.deleteBranch( prefName );

    var nsIMabFile = this.getDir('ProfD');
    nsIMabFile.appendRelativePath( mabFile );
    if (nsIMabFile.exists() && nsIMabFile.isFile())
      nsIMabFile.remove(false);
if(this.dodebug)dump('Old mabfile '+nsIMabFile.path+' removed\n');

    return true;
  } catch (e) {
this.promptService.alert(null, 'ERROR', 'deleteAddressBook('+dir.dirName+')\n'+e);
    return false;
  }
} // deleteAddressBook

/*
 * get DB for addressbook
 */
function getDB(dir) {
  try {
    var curDB;
/*
    var directory = this.abmanager.getDirectory(dir.URI);
    var MDB=directory.QueryInterface(Components.interfaces.nsIAbMDBDirectory);
*/
    if (!(dir instanceof Components.interfaces.nsIAbMDBDirectory)) {
//if(this.dodebug)dump('abs: getDB: addressbook is not a Mozilla addressbook (e.g. LDAP/MACOS)\n');
      return null;
    }
    var MDB=dir.QueryInterface(Components.interfaces.nsIAbMDBDirectory);
//if(this.dodebug)dump('MDB: '+MDB+' '+MDB.database+' '+MDB.databaseFile+'\n');
    curDB=MDB.database; //nsIAddrDatabase
    return curDB;
  } catch(e) {
dump('abs: getDB: '+e+'\n');
    return null;
  }
}

/*
 * remove Listeners from addressbook
 */
/*
function removeListeners(dir, db) {
// remove listeners, db is a nsIAddrDatabase
//not good, won't be reactivated!
  try {
    var mailLists=dir.addressLists;
    var count=mailLists.length;
    for (var i=0; i<count; i++) {
      var list=mailLists.queryElementAt (i, Components.interfaces.nsIAddrDBListener);
      try { db.removeListener(list); }
      catch(e) {dump('db.remListener on list '+e+'\n');} //exception, wenn kein listener existiert
    }
  } catch(e) {
    dump('removeListeners: '+e+'\n');
  }
  try { db.removeListener(dir.QueryInterface(Components.interfaces.nsIAddrDBListener)); }
  catch(e) { dump('db.remListener on dir '+e+'\n'); } //exception, wenn kein listener existiert
if(this.dodebug)dump(count+' listeners removed\n');
}
*/

/*
 * clear an address book
 */
function clearAddressBook(dir)
{
  function deleteLists(ab) {
if(this.dodebug)dump('  delete Lists\n');
    var mdb=ab.QueryInterface(Components.interfaces.nsIAbMDBDirectory);
    var lists=ab.addressLists;
//if(this.dodebug)dump('v');
    var count=lists.length;
    for (var i=count-1; i>=0; i--){ //backwards!
      try {
        var list=lists.queryElementAt(i, Components.interfaces.nsIAbCard);
        var card=list.QueryInterface(Components.interfaces.nsIAbCard);
//if(this.dodebug)dump('  card '+card.primaryEmail+'\n');
//if(this.dodebug)dump('.');
        mdb.removeEmailAddressAt(i);
      } catch(e) {
        var list=lists.queryElementAt(i, Components.interfaces.nsIAbDirectory);
        deleteLists(list);
        list=list.QueryInterface(Components.interfaces.nsIAbMDBDirectory);
        //noscript: list.removeElementsFromAddressList()
//if(this.dodebug)dump('+');
        ab.deleteDirectory(list);
      }
    }
//if(this.dodebug)dump('^');
  }

  function collectCardsToDelete(ab, use_sa) {
if(this.dodebug)dump('  collect cards to delete\n');
    var cardsE=ab.childCards;
    var cards;
    // cardsE is a nsISimpleEnumerator
    if (!use_sa) { //this is the normal case with TB
      cards = Components.classes["@mozilla.org/array;1"].
        createInstance(Components.interfaces.nsIMutableArray);
      while( cardsE.hasMoreElements() ) {
//if(this.dodebug)dump('save card '+card.primaryEmail+'\n');
//if(this.dodebug)dump('.');
        cards.appendElement(cardsE.getNext(), false);
      }
      cards=cards.QueryInterface(Components.interfaces.nsIArray);
    } else { //Postbox wants a nsISupportsArray
      cards = Components.classes["@mozilla.org/supports-array;1"].
        createInstance(Components.interfaces.nsISupportsArray);
      while( cardsE.hasMoreElements() ) {
//if(this.dodebug)dump('save card '+card.primaryEmail+'\n');
//if(this.dodebug)dump('.');
        cards.AppendElement(cardsE.getNext());
      }
    }
    return cards;
  }

  function forcePurgeDeleted(db) {
if(this.dodebug)dump('  purge cards on deleted list\n');
    var count;
    var count=db.deletedCardCount;
    var dList=db.deletedCardList.enumerate();
    while( dList.hasMoreElements() ) {
      var card = dList.getNext().QueryInterface(Components.interfaces.nsIAbCard);
//if(this.dodebug)dump('set lmd on card '+card.primaryEmail+'\n');
      db.setCardValue(card, 'LastModifiedDate', '386d3570' , false);
      count++;
    }
    dList=null;
//if(this.dodebug)dump('lmd set on '+count+' cards, they should vanish soon\n');
  }

  //////////////////
  var curDB=this.getDB(dir);
  //this.removeListeners(dir, curDB);
  deleteLists(dir);
//de.ggbs.abs.dumpDir(dir, 'Lists Deleted:');

  try {
    var cards=collectCardsToDelete(dir, false);
if(this.dodebug)dump('  start deleting cards\n');
    try {
      dir.deleteCards(cards);  // delete the cards from the to-be synced addressbook
    } catch(e) { //try Postbox
if(this.dodebug)dump('Try the Postbox version :-(\n');
      cards=collectCardsToDelete(dir, true);
      dir.deleteCards(cards);  // delete the cards from the to-be synced addressbook
    }
if(this.dodebug)dump('  cards deleted\n');
//de.ggbs.abs.dumpDir(dir, 'Cards deleted:');
    cards = null;
  } catch(e) {
    return 'rab_ve: clear addressbook: '+e;
  }


// The cards are not deleted immediately, they are put on a deletedCards Table
// This table can't be purge programmatically, they are automatically purged
// iff more than 50 Cards are in the table and only cards are purged that are
// older than 6 month :-(((
// Therefore I set the LastModifiedDate attribute on the deleted cards to an
// appropriate date (1.1.2000) so they will be deleted later on
  try  {
    forcePurgeDeleted(curDB);
  } catch(e) {
    return 'rab_ve: forcePurgeDeleted: '+e;
  }

  return '';
} // clearAddressBook

/*
 * Fill Addressbook from mab file
 */
function fillAddressBook(dir, file)
{
  function copyCards(db, ab) {
//if(this.dodebug)dump('  copy cards\n');
    var tCards=db.enumerateCards(ab);
//if(this.dodebug)dump('    copyCards: have enumerated the cards in tempDir\n');
    while(tCards.hasMoreElements()) {
      var tCard = tCards.getNext().QueryInterface(Components.interfaces.nsIAbCard);
      if (!tCard.isMailList) { // only non-maillists
        ab.dropCard(tCard, true); //true is necessary or tb crashes soon after
if(this.dodebug)dump('insert card '+tCard.primaryEmail+'\n');
//if(this.dodebug)dump('.');
      }
    }
    tCards=null;
//if(this.dodebug)dump('\ndone\n');
//de.ggbs.abs.dumpDir(dir, 'Cards copied:');
  }

  function recreateMaillists(ab, db) {
if(this.dodebug)dump('  recreate maillists\n');
    var mailLists=ab.childNodes;
    try  {
      while (mailLists.hasMoreElements()) {
//if(this.dodebug)dump('+');
        var mailList=mailLists.getNext().QueryInterface(Components.interfaces.nsIAbDirectory);
//if(this.dodebug)dump(' recreate '+mailList.dirName+'\n');
        ab.deleteDirectory(mailList);
        ab.addMailList(mailList);
      }
    } catch(e) {
dump('rab_ve: create cards for lists: '+e+'\n');
      return 'rab_ve: create cards for lists: '+e;
    }
    mailLists=null;
//de.ggbs.abs.dumpDir(dir, 'Recreated maillists:');
    return '';
  }

  var curDB=this.getDB(dir);
  //this.removeListeners(dir, curDB);

if(this.dodebug)dump('  open the addressbook database file\n');
  try {
  // open the addressbook database file
    var db=Components.classes["@mozilla.org/addressbook/carddatabase;1"].
      getService(Components.interfaces.nsIAddrDatabase);
    if (!file.isWritable()) { //e.g. a readonly filesystem
      var ds=Services.dirsvc;
      var tdir=ds.get('TmpD',  Components.interfaces.nsIFile)
        .QueryInterface(Components.interfaces.nsILocalFile);
      var tempFile=tdir.clone().QueryInterface(Components.interfaces.nsILocalFile);
      var tempName='abstemp.mab';
      tempFile.appendRelativePath(tempName);
      if (tempFile.exists() && tempFile.isFile()) { tempFile.remove(false); }
        //can't use .copyTo as this also copies the permissions  
        // and i can't change them because that throws an error (at least on windows)
      //file.copyTo( tdir, tempName );  //this copies the permissions too
      //tempFile.permission=/*0666*/0x1b6;
      this.writeMabData(tempFile, this.readMabData(file, ''));
      file=tempFile;
    }
    var tempDB=db.open(file, true, false) // Shows a message and throws error if file is readonly
      .QueryInterface(Components.interfaces.nsIAddrDatabase);
//if(this.dodebug)dump('opened '+file.path+' -> '+tdb+'\n');
  } catch(e) {
    return 'rab_ve: open mabfile: '+file.path+' '+file.permissions+' '+e;
  }

try {
    copyCards(tempDB, dir);
  } catch(e) {
    return 'rab_ve: copy cards: '+e;
  }

  try {
    var msg;
    tempDB.getMailingListsFromDB(dir);
    msg=recreateMaillists(dir, curDB);
    if (msg) this.promptService.alert(null, 'ERROR', msg);
  } catch(e) {
    this.promptService.alert(null, 'ERROR', 'rab_ve: recreate Maillists: '+e);
    return 'rab_ve: recreate Maillists: '+e;
  }
  tempDB.forceClosed();  //!!!
  curDB.close(true);
//  try { if (tempFile) tempFile.remove(false); } catch(e) { return 'rab_ve: remove Tempfile: '+e; }
  return '';
} // fillAddressBook

/*
 * Fill Addressbook from LDIF file
 */
function LDIFAddressBook(dir, file) {
if(this.dodebug)dump('LDIF: started! '+dir.URI+' '+file.path+'\n');

  var selLocIsHome = true;
  var importService = Components.classes["@mozilla.org/import/import-service;1"].
      getService(Components.interfaces.nsIImportService);

    // the index is the 'list-index' of the ldif entry in the 'moduleList' listbox
    // of the import dialog of the addressbook (see chrome\messenger\content\messenger\importDialog.js)
    // 1 up to TB3.2, 2 in TB5 and later
  var index=2;
  var module = importService.GetModule('addressbook', index);

  var observerService=Services.obs;

  var ldifimport = module.GetImportInterface( "addressbook").
      QueryInterface( Components.interfaces.nsIImportGeneric);
  if (ldifimport == null) {
if(this.dodebug)dump('LDIF: No import interface\n');
    var str=Components.classes["@mozilla.org/supports-string;1"].createInstance().
      QueryInterface( Components.interfaces.nsISupportsString);
    str.data='No import interface';
    observerService.notifyObservers(str, "LDIFcomplete", "failed");
    return false;
  }

if(this.dodebug)dump('LDIF: Have ldifimport\n');
  ldifimport.SetData("addressLocation", file);

  var map = ldifimport.GetData( "fieldMap");
  if (map != null) {
if(this.dodebug)dump('LDIF: Not a LDIF file!\n');
    var str=Components.classes["@mozilla.org/supports-string;1"].createInstance().
      QueryInterface( Components.interfaces.nsISupportsString);
    str.data='Not a LDIF file';
    observerService.notifyObservers(str, "LDIFcomplete", "failed");
    return false;
  }
if(this.dodebug)dump('LDIF: have map\n');
  var abURIStr = Components.classes["@mozilla.org/supports-cstring;1"].createInstance().
      QueryInterface( Components.interfaces.nsISupportsCString);
  abURIStr.data=dir.URI;
  ldifimport.SetData("addressDestination", abURIStr);
if(this.dodebug)dump('LDIF: set addressDestination to '+abURIStr+'\n');
//else: created jedesmal ein neues AB mit Namen des LDIF Files

  var cont=LDIFAddressBookCont;
  if (ldifimport.WantsProgress()) {
//if(this.dodebug)dump('wants Progress\n');
    if (ldifimport.BeginImport( cont.successStr, cont.errorStr, selLocIsHome)) {
//if(this.dodebug)dump('start timer\n');
      cont.ldifimport=ldifimport;
      cont.observerService=observerService;
      cont.parent=this;
      var timer = Components.classes["@mozilla.org/timer;1"].
          createInstance(Components.interfaces.nsITimer);
      timer.initWithCallback(cont, 100, Components.interfaces.nsITimer.TYPE_REPEATING_SLACK);
//if(this.dodebug)dump('timer started\n');
    } else {
      observerService.notifyObservers(cont.errorStr, "LDIFcomplete", "failed");
    }
  } else {  //TB6 (or earlier?)
if(this.dodebug)dump('LDIF: not wants Progress\n');
    if (ldifimport.BeginImport( cont.successStr, cont.errorStr, selLocIsHome))
      observerService.notifyObservers(cont.successStr, "LDIFcomplete", "success");
    else
      observerService.notifyObservers(cont.errorStr, "LDIFcomplete", "failed");
  }
  return true;
}
let LDIFAddressBookCont = {
  successStr: Components.classes["@mozilla.org/supports-string;1"].createInstance().
      QueryInterface( Components.interfaces.nsISupportsString),
  errorStr: Components.classes["@mozilla.org/supports-string;1"].createInstance().
      QueryInterface( Components.interfaces.nsISupportsString),
  ldifimport: null,
  observerService: null,
  importSuccess: false,
  parent: null,

  notify: function(timer)
  {
//if(this.dodebug)dump('timer called\n');
    let clear = true;
    let pcnt;

    if (this.ldifimport) {
//if(this.dodebug)dump('continueImport: have interface\n');

      if (!this.ldifimport.ContinueImport()) {
//if(this.dodebug)dump('continueImport: not continue\n');
        this.importSuccess = false;
        timer.cancel();
//if(this.dodebug)dump('LDIF: Import failed: '+this.errorStr+'\n');
        this.observerService.notifyObservers(this.errorStr, "LDIFcomplete", "failed");
      }
      else if ((pcnt = this.ldifimport.GetProgress()) < 100) {
//if(this.dodebug)dump('continueImport: GetProgress='+pcnt+'\n');
        clear = false;
      }
      else {
        timer.cancel();
        this.importSuccess = true;
//if(this.dodebug)dump('LDIF: finished: '+this.successStr+'\n');
//       this.parent.promptService.alert(null, 'INFO', 'LDIF: Import finished: '+this.successStr);
        this.observerService.notifyObservers(this.successStr, "LDIFcomplete", "success");
      }
    }
    if (clear) {
      this.ldifimport = null;
    }
  }
}

/*
 * generate external filename for address book
 */
function externalFilename(dir) {
  var externalFilename;
  var prefName = dir.dirPrefId;
  try { externalFilename = this.mPrefBranch.getCharPref(prefName+'.filename'); }
  catch(e) {
    var uri = dir.URI;
    if (uri!=this.kPersonalAddressbookURI && uri!=this.kCollectedAddressbookURI)
      externalFilename = dir.dirName+'.mab';
    else
      externalFilename = '_tb_'+dir.fileName;
  }
  return externalFilename;
}
function setExternalFilename(dir, externalFilename) {
  var prefName = dir.dirPrefId;
  if (!externalFilename)
    this.mPrefBranch.clearUserPref(prefName+'.filename', externalFilename);
  else
    this.mPrefBranch.setCharPref(prefName+'.filename', externalFilename);
//if(this.dodebug)dump('set external filename '+dir.dirName+'->'+externalFilename+'\n');
  return externalFilename;
}

function addListener() {
  var addrbookSession;
  sessionListener.abs=this;
  sessionListener.dodebug=this.dodebug;
//if(this.dodebug)dump('abs: addListener\n');
  this.abmanager.addAddressBookListener(
    sessionListener,
//      Components.interfaces.nsIAbListener.itemAdded ||
//      Components.interfaces.nsIAbListener.directoryItemRemoved ||
//      Components.interfaces.nsIAbListener.directoryRemoved ||
//      Components.interfaces.nsIAbListener.itemChanged ||
      Components.interfaces.nsIAbListener.all);
}

function removeListener() {
//if(this.dodebug)dump('abs: removeListener\n');
  this.abmanager.removeAddressBookListener(sessionListener);
}

var sessionListener = {
  abs: null,
  dodebug: false,

  // these functions are called for the main dir and every list (which has been shown
  // in the addressbookwindows) even if item are not part of the list
//After book has been downloaded vie replace by entry, these functions
//are no longer called for dir (only for lists)
  onItemAdded: function(parentDir, item) {
    try {
      var dir=parentDir.QueryInterface(Components.interfaces.nsIAbDirectory);
//if(this.dodebug)dump('dir.URI: '+dir.URI+'\n'); //moz-abmdbdirectory://abook.mab
      if (item instanceof Components.interfaces.nsIAbCard) {
        var card=item.QueryInterface(Components.interfaces.nsIAbCard);
        //card might be a list
        var now=Math.round(Date.now()/1000);
if(this.dodebug)dump('abs: item added parent='+dir.dirName+' email='+card.primaryEmail+'\n');
//if(this.dodebug)dump('     lmd of dir,card set from '+dir.lastModifiedDate+','+card.getCardValue('LastModifiedDate')+' to '+now+'\n');
if(this.dodebug)dump('     lmd of dir,card set from '+dir.lastModifiedDate+','+card.getProperty('LastModifiedDate','???')+' to '+now+'\n');
        dir.lastModifiedDate=now;
        //card.setProperty('LastModifiedDate',now);
      }
    } catch(e) { dump('abs Listener (item added '+item+'): '+e+'\n'); }
  },
  onItemRemoved: function(parentDir, item) {
    // if a list is removed, this is called twice, with a nsIAbCard than with nsIAbDirectory
    try {
      var now=Math.round(Date.now()/1000);
if(this.dodebug)dump('abs: remove item time='+now+'\n');
      var dir=parentDir.QueryInterface(Components.interfaces.nsIAbDirectory);
      if (item instanceof Components.interfaces.nsIAbCard) {
var card=item.QueryInterface(Components.interfaces.nsIAbCard);
if(this.dodebug)dump('abs: card removed parent='+dir.dirName+' email='+card.primaryEmail+'\n');
if(this.dodebug)dump('     lmd of dir set from '+dir.lastModifiedDate+' to '+now+'\n');
        dir.lastModifiedDate=now;
      } else if (item instanceof Components.interfaces.nsIAbDirectory) {
        var list=item.QueryInterface(Components.interfaces.nsIAbDirectory);
        if (!list.dirName) {
if(this.dodebug)dump('abs: book removed\n');  // complete book has been removed
        } else {
if(this.dodebug)dump('abs: list removed parent='+dir.dirName+' list='+list.dirName+'\n');
          // hhm, wie unterscheide ich, ob es nur ein löschen der Liste
          // ist, oder ein löschen des ganzen Buches?
          // Vielleicht ist es auch egal?
if(this.dodebug)dump('     lmd of dir set from '+dir.lastModifiedDate+' to '+now+'\n');
          dir.lastModifiedDate=now;
        }
      } else {
if(this.dodebug)dump('abs: ??? removed parent='+dir.dirName+'\n');
      }
    } catch(e) { dump('abs Listener (item deleted): '+e+'\n'); }
  },
  onItemPropertyChanged: function(item, property, oldValue, newValue) {
    try{
      var now=Math.round(Date.now()/1000);
if(this.dodebug)dump('abs: property change time='+now+'\n');
      //property, oldValue and newValue are always null
      if (item instanceof Components.interfaces.nsIAbCard) {
        var card=item.QueryInterface(Components.interfaces.nsIAbCard);
        //property and values always null
if(this.dodebug)dump('card changed '+card.primaryEmail+'\n');
//if(this.dodebug)dump('  directoryId: '+card.directoryId+' localId: '+card.localId+'\n');  //ldap_2.servers.pab&Personal
if(this.dodebug)dump('     lmd of card is '+card.getProperty('LastModifiedDate','???')+'\n');
        // lastModifiedDate of card has changed, but not for the directoy :-(
        var dir=this.abs.getAddressBook(card.directoryId.replace(/&.*$/,''));
        if (dir) {
if(this.dodebug)dump('     lmd of dir set from '+dir.lastModifiedDate+' to '+now+'\n');
          dir.lastModifiedDate=now;
        } else dump('ABS: Card change, but no directory found for '+card.directoryId+'\n');
      } else if (item instanceof Components.interfaces.nsIAbDirectory) {
        var dir=item.QueryInterface(Components.interfaces.nsIAbDirectory);
        //property='DirName'
if(this.dodebug)dump('dir changed '+dir.dirName+' '+property+': '+oldValue+'->'+newValue+'\n');
        if (dir.isMailList) {
if(this.dodebug)dump('  dir is MailList: '+dir.URI+'\n');   //moz-abmdbdirectory://abook.mab/MailList3
          dir=this.abs.abmanager.getDirectory(dir.URI.replace(/\/[^\/]*$/,''));
if(this.dodebug)dump('     lmd of dir set from '+dir.lastModifiedDate+' to '+now+'\n');
          dir.lastModifiedDate=now;
        }
      } else {
if(this.dodebug)dump('abs: ??? changed item='+item+'\n');
      }
    } catch(e) { dump('abs Listener (item changed): '+e+'\n'); }
  }
};

function getDir(special) {
/*
  var profileFolder = directoryService.get(special,  Components.interfaces.nsIFile);
  var profileDir = profileFolder.QueryInterface(Components.interfaces.nsILocalFile);
  var nsIMabFile = profileDir.clone().QueryInterface(Components.interfaces.nsILocalFile);
*/
  var dir = this.directoryService.get(special,  Components.interfaces.nsIFile)
    .QueryInterface(Components.interfaces.nsILocalFile);
//  return dir.clone().QueryInterface(Components.interfaces.nsILocalFile);
  return dir;
}

/*
 *  Read a mab file and return the data
 */
function readMabData(nsIMabFile, addLastMod)
{
  if (!nsIMabFile.exists()) return '';

  //IOUtils.js: loadFileToString(file)
  var fs = Components.classes['@mozilla.org/network/file-input-stream;1']
          .createInstance(Components.interfaces.nsIFileInputStream);
  var ss = Components.classes['@mozilla.org/scriptableinputstream;1']
          .createInstance(Components.interfaces.nsIScriptableInputStream);
  var str='';
  try{
    fs.init(nsIMabFile, 0x01, 0, false);
    ss.init(fs);
    str = ss.read(-1);
    ss.close();
    fs.close();
  } catch(e){ this.promptService.alert(null, 'readMabData', 'readMabData:\n'+e); }
//    return str;
  if (addLastMod) {
    var lmt=new Date(nsIMabFile.lastModifiedTime);
    return str.replace(/-->[^\r\n]*/,'--> '+lmt);//append to the // <!-- <mdb:mork:z v="1.4"/> -->
  } else
    return str;
}

/*
 *  Save Data to a mab File
 */
function writeMabData(nsIMabFile, abData)
{
  try {
    if (nsIMabFile.exists() && nsIMabFile.isFile()) nsIMabFile.remove(false);
    var strm = Components.classes["@mozilla.org/network/file-output-stream;1"].
      createInstance(Components.interfaces.nsIFileOutputStream);
    strm.QueryInterface(Components.interfaces.nsIOutputStream);
    strm.QueryInterface(Components.interfaces.nsISeekableStream);
    strm.init( nsIMabFile, 0x20|0x02|0x08, /*0600*/0x180, 0 );
      // 0x02=PR_WRONLY, 0x08=PR_CREATE_FILE, 0x20=PR_TRUNCATE
    strm.write( abData, abData.length );
    strm.flush();
    strm.close();
  } catch(e) {
      this.promptService.alert(null, 'writeMabData', 'File write error of '+nsIMabFile.target+'\n'+e);
      return null;
  }
  return nsIMabFile;
}

/*
 *  This is, where every up-/download starts!
 */
function showPopup(direction, singlebook, type, force)
{
  //singlebook: a mabFile or null
  //type: start, final, auto, manual
  //force: if true, upload mabs, even if not changed
  if (this.synctype=='none') return;
  if (direction=='upload' && this.noupload) return;
  if (!singlebook && this.syncedAddressbooks((direction=='download')?'down':'up',null).Count()==0)
    return;  //nothing to do

//TODO: possible race condition!
  if (this.state!='') {  // we are already doing an up-/download
    if (type=='manual')
      this.promptService.alert(null, 'showPopup', this.strings.GetStringFromName('inprogress'));
    return;
  }

  this.usesPopup=
    type=='final' ||     //always at propram exit
    (type=='start' && !this.hidepopups && !this.hideallpopups) ||
    (type!='final' && type!='start' && !this.hideallpopups);

  this.state=direction;

this.debug('showPopup usesPopup='+this.usesPopup);
  if (!this.usesPopup) {
    var windowsEnum = this.wMediator.getEnumerator(null);  //enumerate all windows //"mail:3pane");
    while (windowsEnum.hasMoreElements()) {
      var w=windowsEnum.getNext();
      if (w.de.ggbs.abs.upload) break;
      w=null;
    }
    if (!w) {
this.debug('showPopup: no Window!');
      this.promptService.alert(null, 'showPopup', 'no window to perform up-/download!');
      this.state='';
    } else{
this.debug('showPopup: show');
      if (direction=='download') w.de.ggbs.abs.download(singlebook, force);    //in download.js
      else                       w.de.ggbs.abs.upload(singlebook, force);      //in upload.js
    }
  } else {
/*
    var w=window.openDialog("chrome://addressbookssync/content/"+direction+".xul","_blank",
      'chrome,resizable,titlebar=no', singlebook, force);
*/
    var args = Components.classes["@mozilla.org/supports-array;1"]
                    .createInstance(Components.interfaces.nsISupportsArray);
    var s=Components.classes["@mozilla.org/supports-string;1"]
        .createInstance(Components.interfaces.nsISupportsString);
    s.data=singlebook;
    args.AppendElement(s);
    var f=Components.classes["@mozilla.org/supports-string;1"]
        .createInstance(Components.interfaces.nsISupportsString);
    f.data=force?'true':'false';
    args.AppendElement(f);
    var modal='';
    if (this.exiting && this.synctype=='imap') {
      modal=',modal';
/*
      if (this.lastWindow) this.lastWindow.minimize();
      var windowsEnum = this.wMediator.getEnumerator(null);  //enumerate all windows //"mail:3pane");
      while (windowsEnum.hasMoreElements()) {
        var w=windowsEnum.getNext();
        w.minimize();
      }
*/
      this.debug('showPopup: openmodal');
    }
    var w=this.wWatcher.openWindow(null, "chrome://addressbookssync/content/"+direction+".xul",
      '_blank', 'chrome,resizable,titlebar=no'+modal, args);
    //if adding more arguments: update in call to upload() and download() above
    //    AND in upload.xul and download.xul!!
this.debug('showPopup: opened '+w);
  }
}

function debug(text) {
  if (!this.dodebug) return;
  var d=new Date();
  var s=d.toLocaleFormat('%d.%b.%Y %H:%M:%S');
  text=s+': '+text+'\n';
  var flags=0x10|0x08|0x02;
        // 0x02=PR_WRONLY, 0x10=PR_APPEND, 0x08=PR_CREATE_FILE
  try {
    var logFile = this.getDir('TmpD');
    logFile.appendRelativePath( 'AddressbookSynchronizer.log' );
    var strm = Components.classes["@mozilla.org/network/file-output-stream;1"].
      createInstance(Components.interfaces.nsIFileOutputStream);
    strm.QueryInterface(Components.interfaces.nsIOutputStream);
    strm.QueryInterface(Components.interfaces.nsISeekableStream);
    strm.init( logFile, flags, /*0600*/0x180, 0 );
    strm.write( text, text.length );
    strm.flush();
    strm.close();
  } catch(e) { console.logStringMessage('AddressbooksSynchronizer: '+'File write:\n'+e+'\n'); }
}
